#!/usr/bin/python3
# Closed-box tests of netplan CLI. These are run during "make check" and don't
# touch the system configuration at all.
#
# Copyright (C) 2023 Canonical, Ltd.
# Authors: Danilo Egea Gondolfo <danilo.egea.gondolfo@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import tempfile
import unittest

from unittest.mock import Mock
from netplan.netdef import NetplanRoute
from netplan_cli.cli.state import Interface, NetplanConfigState, SystemConfigState
from netplan_cli.cli.state_diff import DiffJSONEncoder, NetplanDiffState


class TestNetplanDiff(unittest.TestCase):
    '''Test netplan state NetplanDiffState class'''

    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory(prefix='netplan_')
        self.file = '90-netplan.yaml'
        self.path = os.path.join(self.workdir.name, 'etc', 'netplan', self.file)
        os.makedirs(os.path.join(self.workdir.name, 'etc', 'netplan'))

        self.diff_state = NetplanDiffState(Mock(spec=SystemConfigState), Mock(spec=NetplanConfigState))
        self.diff_state.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}

    def test_get_full_state(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      dhcp4: true
      dhcp6: false''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)
        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
                'type': 'ethernet',
                'addresses': [
                    {
                        '1.2.3.4': {
                            'prefix': 24,
                            'flags': ['dhcp'],
                        }
                    },
                ],
            }
        }
        system_state.interface_list = []
        diff_state = NetplanDiffState(system_state, netplan_state)

        full_state = diff_state.get_full_state()
        expected = {
            'interfaces': {
                'eth0': {
                    'system_state': {
                        'type': 'ethernet',
                        'addresses': {
                            '1.2.3.4/24': {
                                'flags': ['dhcp']
                            }
                        },
                        'id': 'mynic',
                        'index': 2,
                    },
                    'netplan_state': {
                        'id': 'mynic',
                        'type': 'ethernet',
                        'dhcp4': True,
                        'dhcp6': False,
                        'link_local': ['ipv6'],
                    }
                }
            }
        }

        self.assertDictEqual(full_state, expected)

    def test_get_netplan_interfaces(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      dhcp4: false
      dhcp6: false
      macaddress: aa:bb:cc:dd:ee:ff
      routes:
        - to: default
          via: 1.2.3.4
      nameservers:
        addresses:
          - 1.1.1.1
          - 2.2.2.2
        search:
          - mydomain.local
      addresses:
        - 192.168.0.2/24:
            label: myip
            lifetime: forever
        - 192.168.0.1/24''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)
        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
            }
        }
        system_state.interface_list = []
        diff_state = NetplanDiffState(system_state, netplan_state)

        interfaces = diff_state._get_netplan_interfaces()
        expected = {
            'eth0': {
                'netplan_state': {
                    'id': 'mynic',
                    'addresses': {
                        '192.168.0.1/24': {
                            'flags': {}
                        },
                        '192.168.0.2/24': {
                            'flags': {'label': 'myip', 'lifetime': 'forever'},
                        }
                    },
                    'dhcp4': False,
                    'dhcp6': False,
                    'link_local': ['ipv6'],
                    'nameservers_addresses': ['1.1.1.1', '2.2.2.2'],
                    'nameservers_search': ['mydomain.local'],
                    'macaddress': 'aa:bb:cc:dd:ee:ff',
                    'type': 'ethernet',
                    'routes': [NetplanRoute(to='default', via='1.2.3.4', family=2)],
                }
            }
        }
        self.assertDictEqual(interfaces, expected)

    def test_get_netplan_interfaces_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    # not matching any physical device
    myeths:
      match:
        name: eth*
    mynics:
      dhcp4: false
      dhcp6: false
      match:
        name: enp0*''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)
        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'enp0s3': {
                'name': 'enp0s3',
                'id': 'mynics',
                'index': 2,
            },
            'enp0s4': {
                'name': 'enp0s4',
                'id': 'mynics',
                'index': 3,
            },
            'enp0s5': {
                'name': 'enp0s5',
                'id': 'mynics',
                'index': 4,
            }
        }
        system_state.interface_list = []
        diff_state = NetplanDiffState(system_state, netplan_state)

        interfaces = diff_state._get_netplan_interfaces()
        expected = {
            'enp0s3': {
                'netplan_state': {
                    'id': 'mynics',
                    'dhcp4': False,
                    'dhcp6': False,
                    'type': 'ethernet',
                    'link_local': ['ipv6'],
                }
            },
            'enp0s4': {
                'netplan_state': {
                    'id': 'mynics',
                    'dhcp4': False,
                    'dhcp6': False,
                    'type': 'ethernet',
                    'link_local': ['ipv6'],
                }
            },
            'enp0s5': {
                'netplan_state': {
                    'id': 'mynics',
                    'dhcp4': False,
                    'dhcp6': False,
                    'type': 'ethernet',
                    'link_local': ['ipv6'],
                },
            },
            'myeths': {
                'netplan_state': {
                    'id': 'myeths',
                    'dhcp4': False,
                    'dhcp6': False,
                    'type': 'ethernet',
                    'link_local': ['ipv6'],
                }
            }
        }
        self.assertDictEqual(interfaces, expected)

    def test_get_system_interfaces(self):
        system_state = Mock(spec=SystemConfigState)
        netplan_state = Mock(spec=NetplanConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'type': 'ethernet',
                'index': 2,
                'addresses': [
                    {
                        '1.2.3.4': {
                            'prefix': 24,
                            'flags': [],
                        }
                    },
                ],
                'dns_addresses': ['1.1.1.1', '2.2.2.2'],
                'dns_search': ['mydomain.local'],
                'routes': [
                    {
                        'to': 'default',
                        'via': '192.168.5.1',
                        'from': '192.168.5.122',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    }
                ],
                'macaddress': 'aa:bb:cc:dd:ee:ff',
            }
        }
        system_state.interface_list = []

        diff_state = NetplanDiffState(system_state, netplan_state)
        diff_state.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        interfaces = diff_state._get_system_interfaces()
        expected = {
            'eth0': {
                'system_state': {
                    'type': 'ethernet',
                    'id': 'mynic',
                    'index': 2,
                    'addresses': {
                        '1.2.3.4/24': {
                            'flags': []
                        }
                    },
                    'nameservers_addresses': ['1.1.1.1', '2.2.2.2'],
                    'nameservers_search': ['mydomain.local'],
                    'routes': [
                        NetplanRoute(to='default',
                                     via='192.168.5.1',
                                     from_addr='192.168.5.122',
                                     type='unicast',
                                     scope='global',
                                     protocol='kernel',
                                     table=254,
                                     family=2,
                                     metric=100)
                    ],
                    'macaddress': 'aa:bb:cc:dd:ee:ff'
                },
            }
        }
        self.assertDictEqual(interfaces, expected)

    def test_diff_default_table_names_to_number(self):
        self.assertEqual(self.diff_state._default_route_tables_name_to_number('main'), 254)
        self.assertEqual(self.diff_state._default_route_tables_name_to_number('default'), 253)
        self.assertEqual(self.diff_state._default_route_tables_name_to_number('local'), 255)
        self.assertEqual(self.diff_state._default_route_tables_name_to_number('1000'), 1000)
        self.assertEqual(self.diff_state._default_route_tables_name_to_number('blah'), 0)

    def test__system_route_to_netplan_empty_input(self):
        route = self.diff_state._system_route_to_netplan({})
        expected = NetplanRoute()
        self.assertEqual(route, expected)

    def test__system_route_to_netplan(self):
        route = {
            'to': 'default',
            'via': '192.168.5.1',
            'from': '192.168.5.122',
            'metric': 100,
            'type': 'unicast',
            'scope': 'global',
            'protocol': 'kernel',
            'family': 2,
            'table': 'main'
        }

        netplan_route = self.diff_state._system_route_to_netplan(route)
        expected = NetplanRoute(to='default', via='192.168.5.1', from_addr='192.168.5.122',
                                metric=100, type='unicast', scope='global', protocol='kernel',
                                family=2, table=254)
        self.assertEqual(netplan_route, expected)

    def test_diff_missing_netplan_interface(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets: {}''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'index': 2,
            },
            'lo': {
                'name': 'lo',
                'index': 1,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'lo'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('missing_interfaces_netplan', [])
        self.assertIn('eth0', missing)
        # lo is included
        self.assertIn('lo', missing)

    def test_diff_missing_system_interface(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
    eth1: {}''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            }
        }
        interface = Mock(spec=Interface)
        interface.name = 'eth0'
        interface.netdef_id = 'eth0'
        interface.data_sources = {}
        system_state.interface_list = [interface]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()
        missing = diff_data.get('missing_interfaces_system', [])
        self.assertIn('eth1', missing)

    def test_diff_missing_system_interface_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynics:
      dhcp4: false
      match:
        name: eth*''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'enp0s1': {
                'name': 'enp0s1',
                'id': 'enp0s1',
                'index': 2,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'enp0s1'
        interface1.netdef_id = 'enp0s1'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()
        missing = diff_data.get('missing_interfaces_system', [])
        self.assertIn('mynics', missing)

    def test_diff_not_missing_system_interface_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynics:
      dhcp4: false
      match:
        name: eth*''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynics',
                'index': 2,
            },
            'enp0s1': {
                'name': 'enp0s1',
                'id': 'enp0s1',
                'index': 3,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynics'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'enp0s1'
        interface2.netdef_id = 'enp0s1'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()
        missing = diff_data.get('missing_interfaces_system', {})
        self.assertDictEqual(missing, {})

    def test__get_comparable_interfaces_empty(self):
        res = self.diff_state._get_comparable_interfaces({})
        self.assertDictEqual(res, {})

    def test__get_comparable_interfaces(self):
        input = {
            'eth0': {
                'system_state': {
                    'id': 'eth0'
                },
                'netplan_state': {}
            },
            'eth1': {
                'netplan_state': {}
            },
            'eth2': {
                'system_state': {}
            },
            'eth3': {
                'system_state': {},
                'netplan_state': {}
            },
        }
        res = self.diff_state._get_comparable_interfaces(input)
        self.assertDictEqual(res, {'eth0': {'netplan_state': {}, 'system_state': {'id': 'eth0'}}})

    def test__compress_ipv6_address_with_prefix(self):
        self.assertEqual(self.diff_state._compress_ipv6_address('a:b:c:0:0:0::d/64'), 'a:b:c::d/64')

    def test__compress_ipv6_address_without_prefix(self):
        self.assertEqual(self.diff_state._compress_ipv6_address('a:b:c:0:0:0::d'), 'a:b:c::d')

    def test__compress_ipv6_address_ipv4_with_prefix(self):
        self.assertEqual(self.diff_state._compress_ipv6_address('192.168.0.1/24'), '192.168.0.1/24')

    def test__compress_ipv6_address_ipv4_without_prefix(self):
        self.assertEqual(self.diff_state._compress_ipv6_address('192.168.0.1'), '192.168.0.1')

    def test__compress_ipv6_address_not_an_ip(self):
        self.assertEqual(self.diff_state._compress_ipv6_address('default'), 'default')

    def test__normalize_ip_addresses(self):
        ips = {'abcd:0:0:0::1/64', '1:2:0:0::123', '1.2.3.4/24', '1.2.3.4'}
        expected = {'abcd::1/64', '1:2::123', '1.2.3.4/24', '1.2.3.4'}
        result = self.diff_state._normalize_ip_addresses(ips)
        self.assertSetEqual(expected, result)

    def test_diff_missing_system_address(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: false
      dhcp6: false
      addresses:
        - 192.168.0.2/24:
            label: myip
            lifetime: forever
        - 192.168.0.1/24''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_addresses', [])
        self.assertIn('192.168.0.1/24', missing)
        self.assertIn('192.168.0.2/24', missing)

    def test_diff_missing_system_address_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: "eth*"
      dhcp4: false
      dhcp6: false
      addresses:
        - 192.168.0.2/24:
            label: myip
            lifetime: forever
        - 192.168.0.1/24''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_addresses', [])
        self.assertIn('192.168.0.1/24', missing)
        self.assertIn('192.168.0.2/24', missing)

    def test_diff_dhcp_addresses_are_filtered_out(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      dhcp6: true''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'addresses': [
                    {'192.168.0.1': {'prefix': 24, 'flags': ['dhcp']}},
                    {'192.168.254.1': {'prefix': 24, 'flags': ['dhcp']}},
                    {'abcd:1234::1': {'prefix': 64, 'flags': ['dhcp']}}
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        self.assertEqual(missing, [])

    def test_diff_ra_addresses_are_filtered_out_with_accept_ra(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      accept-ra: true''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            "netplan-global-state": {},
            "eth0": {
                "index": 4,
                "id": "eth0",
                "addresses": [
                    {"10.86.126.253": {"prefix": 24, "flags": ["dynamic", "dhcp"]}},
                    {"fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d": {"prefix": 64, "flags": ["ra"]}},
                    {"fe80::8840:4dff:fea6:655d": {"prefix": 64, "flags": ["link"]}}
                ],
                "dns_addresses": [
                    "10.86.126.1",
                    "fd42:bc43:e20e:8cf7::1",
                    "fe80::216:3eff:feab:beb9"
                ],
                "routes": [
                    {
                        "to": "default",
                        "family": 2,
                        "via": "10.86.126.1",
                        "from": "10.86.126.253",
                        "metric": 100,
                        "type": "unicast",
                        "scope": "global",
                        "protocol": "dhcp",
                        "table": "main"
                    },
                    {
                        "to": "10.86.126.0/24",
                        "family": 2,
                        "from": "10.86.126.253",
                        "metric": 100,
                        "type": "unicast",
                        "scope": "link",
                        "protocol": "kernel",
                        "table": "main"
                    },
                    {
                        "to": "10.86.126.1",
                        "family": 2,
                        "from": "10.86.126.253",
                        "metric": 100,
                        "type": "unicast",
                        "scope": "link",
                        "protocol": "dhcp",
                        "table": "main"
                    },
                    {
                        "to": "10.86.126.253",
                        "family": 2,
                        "from": "10.86.126.253",
                        "type": "local",
                        "scope": "host",
                        "protocol": "kernel",
                        "table": "local"
                    },
                    {
                        "to": "10.86.126.255",
                        "family": 2,
                        "from": "10.86.126.253",
                        "type": "broadcast",
                        "scope": "link",
                        "protocol": "kernel",
                        "table": "local"
                    },
                    {
                        "to": "fd42:bc43:e20e:8cf7::/64",
                        "family": 10,
                        "metric": 100,
                        "type": "unicast",
                        "scope": "global",
                        "protocol": "ra",
                        "table": "main"
                    },
                    {
                        "to": "fe80::/64",
                        "family": 10,
                        "metric": 256,
                        "type": "unicast",
                        "scope": "global",
                        "protocol": "kernel",
                        "table": "main"
                    },
                    {
                        "to": "default",
                        "family": 10,
                        "via": "fe80::216:3eff:feab:beb9",
                        "metric": 100,
                        "type": "unicast",
                        "scope": "global",
                        "protocol": "ra",
                        "table": "main"
                    },
                    {
                        "to": "fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d",
                        "family": 10,
                        "type": "local",
                        "scope": "global",
                        "protocol": "kernel",
                        "table": "local"
                    },
                    {
                        "to": "fe80::8840:4dff:fea6:655d",
                        "family": 10,
                        "type": "local",
                        "scope": "global",
                        "protocol": "kernel",
                        "table": "local"
                    },
                    {
                        "to": "ff00::/8",
                        "family": 10,
                        "metric": 256,
                        "type": "multicast",
                        "scope": "global",
                        "protocol": "kernel",
                        "table": "local"
                    }
                ],
            }
        }

        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        self.assertEqual(missing, [])

    def test_diff_ra_addresses_are_filtered_out_with_accept_ra_false(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      dhcp6: true
      accept-ra: false''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'index': 4,
                'id': 'eth0',
                'addresses': [
                    {'10.86.126.253': {'prefix': 24, 'flags': ['dynamic', 'dhcp']}},
                    {'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d': {'prefix': 64, 'flags': ['ra']}},
                    {'fe80::8840:4dff:fea6:655d': {'prefix': 64, 'flags': ['link']}}
                ],
                'dns_addresses': [
                    '10.86.126.1',
                    'fd42:bc43:e20e:8cf7::1',
                    'fe80::216:3eff:feab:beb9',
                    'abcd::123',
                ],
                'routes': [
                    {
                        'to': 'default',
                        'family': 2,
                        'via': '10.86.126.1',
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'dhcp',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.0/24',
                        'family': 2,
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'link',
                        'protocol': 'kernel',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.1',
                        'family': 2,
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'link',
                        'protocol': 'dhcp',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.253',
                        'family': 2,
                        'from': '10.86.126.253',
                        'type': 'local',
                        'scope': 'host',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': '10.86.126.255',
                        'family': 2,
                        'from': '10.86.126.253',
                        'type': 'broadcast',
                        'scope': 'link',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'fd42:bc43:e20e:8cf7::/64',
                        'family': 10,
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'ra',
                        'table': 'main'
                    },
                    {
                        'to': 'fe80::/64',
                        'family': 10,
                        'metric': 256,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'main'
                    },
                    {
                        'to': 'default',
                        'family': 10,
                        'via': 'fe80::216:3eff:feab:beb9',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'ra',
                        'table': 'main'
                    },
                    {
                        'to': 'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d',
                        'family': 10,
                        'type': 'local',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'fe80::8840:4dff:fea6:655d',
                        'family': 10,
                        'type': 'local',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'ff00::/8',
                        'family': 10,
                        'metric': 256,
                        'type': 'multicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    }
                ],
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {
            'addresses': {
                'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d': 'NDisc'
            },
            'dns': {
                '10.86.126.1': 'DHCPv4',
                'fd42:bc43:e20e:8cf7::1': 'DHCPv6',
                'fe80::216:3eff:feab:beb9': 'NDisc',
                'abcd::123': 'DHCPv6',
            }
        }
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()
        expected = {
            'missing_addresses': ['fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d/64'],
            'missing_nameservers_addresses':
            ['abcd::123', 'fd42:bc43:e20e:8cf7::1', 'fe80::216:3eff:feab:beb9'],
            'missing_routes': [
                NetplanRoute(to='default',
                             via='fe80::216:3eff:feab:beb9',
                             from_addr=None,
                             type='unicast',
                             scope='global',
                             protocol='ra',
                             table=254,
                             family=10,
                             metric=100,
                             mtubytes=0,
                             congestion_window=0,
                             advertised_receive_window=0,
                             onlink=False),
                NetplanRoute(to='fd42:bc43:e20e:8cf7::/64',
                             via=None,
                             from_addr=None,
                             type='unicast',
                             scope='global',
                             protocol='ra',
                             table=254,
                             family=10,
                             metric=100,
                             mtubytes=0,
                             congestion_window=0,
                             advertised_receive_window=0,
                             onlink=False)]}

        netplan_state = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {})
        missing_addresses = netplan_state.get('missing_addresses', [])
        missing_ns = netplan_state.get('missing_nameservers_addresses', [])
        missing_routes = netplan_state.get('missing_routes', [])
        self.assertListEqual(missing_addresses, expected.get('missing_addresses', []))
        self.assertListEqual(sorted(missing_ns), sorted(expected.get('missing_nameservers_addresses', [])))
        self.assertListEqual(missing_routes, expected.get('missing_routes', []))

    def test_diff_dhcpv6_addresses_are_not_filtered_out_with_accept_ra_true_and_dhcp6_false(self):

        # DHCPv6 will be triggered in networkd by RA regardless if dhcp6 is false.
        # See systemd.network(5)
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      dhcp6: false
      accept-ra: true''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'index': 4,
                'id': 'eth0',
                'addresses': [
                    {'10.86.126.253': {'prefix': 24, 'flags': ['dynamic', 'dhcp']}},
                    {'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d': {'prefix': 64, 'flags': ['ra']}},
                    {'fe80::8840:4dff:fea6:655d': {'prefix': 64, 'flags': ['link']}}
                ],
                'dns_addresses': [
                    '10.86.126.1',
                    'fd42:bc43:e20e:8cf7::1',
                    'fe80::216:3eff:feab:beb9',
                    'abcd::123',
                ],
                'routes': [
                    {
                        'to': 'default',
                        'family': 2,
                        'via': '10.86.126.1',
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'dhcp',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.0/24',
                        'family': 2,
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'link',
                        'protocol': 'kernel',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.1',
                        'family': 2,
                        'from': '10.86.126.253',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'link',
                        'protocol': 'dhcp',
                        'table': 'main'
                    },
                    {
                        'to': '10.86.126.253',
                        'family': 2,
                        'from': '10.86.126.253',
                        'type': 'local',
                        'scope': 'host',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': '10.86.126.255',
                        'family': 2,
                        'from': '10.86.126.253',
                        'type': 'broadcast',
                        'scope': 'link',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'fd42:bc43:e20e:8cf7::/64',
                        'family': 10,
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'ra',
                        'table': 'main'
                    },
                    {
                        'to': 'fe80::/64',
                        'family': 10,
                        'metric': 256,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'main'
                    },
                    {
                        'to': 'default',
                        'family': 10,
                        'via': 'fe80::216:3eff:feab:beb9',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'ra',
                        'table': 'main'
                    },
                    {
                        'to': 'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d',
                        'family': 10,
                        'type': 'local',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'fe80::8840:4dff:fea6:655d',
                        'family': 10,
                        'type': 'local',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    },
                    {
                        'to': 'ff00::/8',
                        'family': 10,
                        'metric': 256,
                        'type': 'multicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'table': 'local'
                    }
                ],
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {
            'addresses': {
                'fd42:bc43:e20e:8cf7:8840:4dff:fea6:655d': 'NDisc'
            },
            'dns': {
                '10.86.126.1': 'DHCPv4',
                'fd42:bc43:e20e:8cf7::1': 'DHCPv6',
                'fe80::216:3eff:feab:beb9': 'NDisc',
                'abcd::123': 'DHCPv6',
            }
        }
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()
        expected = {
            'missing_addresses': [],
            'missing_nameservers_addresses': [],
            'missing_routes': []}

        netplan_state = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {})
        missing_addresses = netplan_state.get('missing_addresses', [])
        missing_ns = netplan_state.get('missing_nameservers_addresses', [])
        missing_routes = netplan_state.get('missing_routes', [])
        self.assertListEqual(missing_addresses, expected.get('missing_addresses', []))
        self.assertListEqual(sorted(missing_ns), sorted(expected.get('missing_nameservers_addresses', [])))
        self.assertListEqual(missing_routes, expected.get('missing_routes', []))

    def test_diff_missing_netplan_address(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: false
      dhcp6: false
      addresses:
        - 192.168.0.1/24''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'addresses': [
                    {'192.168.0.1': {'prefix': 24}},
                    {'192.168.254.1': {'prefix': 24}}
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)

        diff_data = diff.get_diff()
        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        self.assertIn('192.168.254.1/24', missing)

        diff_data = diff.get_diff('eth0')
        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        self.assertIn('192.168.254.1/24', missing)

        # eth1 does not exist
        diff_data = diff.get_diff('eth1')
        self.assertDictEqual(diff_data['interfaces'], {})

    def test_diff_addresses_compressed_ipv6(self):
        ''' Check if IPv6 address will not mismatch due to their representation'''
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: false
      dhcp6: false
      addresses:
        - 1:2:3:0:0:0::123/64''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'addresses': [
                    {'1:2:3::123': {'prefix': 64}},
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_addresses', [])
        self.assertListEqual([], missing_netplan)
        self.assertListEqual([], missing_system)

    def test_diff_missing_system_dhcp_addresses(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      dhcp6: true''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        dhcp4 = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_dhcp4_address')
        dhcp6 = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_dhcp6_address')
        self.assertTrue(dhcp4)
        self.assertTrue(dhcp6)

    def test_diff_link_local_addresses(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: false
      dhcp6: false
      link-local: []''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'addresses': [
                    {'169.254.65.85': {'prefix': 16, 'flags': ['link']}},
                    {'fe80::4e7:f4ff:fe6e:c917': {'prefix': 64, 'flags': ['link']}},
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)

        diff_data = diff.get_diff()
        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_addresses', [])
        self.assertIn('169.254.65.85/16', missing)
        self.assertIn('fe80::4e7:f4ff:fe6e:c917/64', missing)

    def test_diff_missing_system_nameservers(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      nameservers:
        addresses:
          - 1.2.3.4
          - 4.3.2.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_nameservers_addresses', [])
        self.assertIn('1.2.3.4', missing)
        self.assertIn('4.3.2.1', missing)

    def test_diff_missing_netplan_nameservers(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'dns_addresses': ['1.2.3.4', '4.3.2.1'],
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        state = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {})
        missing = state.get('missing_nameservers_addresses', [])
        self.assertIn('1.2.3.4', missing)
        self.assertIn('4.3.2.1', missing)

    def test_diff_missing_system_nameservers_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      nameservers:
        addresses:
          - 1.2.3.4
          - 4.3.2.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_nameservers_addresses', [])
        self.assertIn('1.2.3.4', missing)
        self.assertIn('4.3.2.1', missing)

    def test_diff_missing_system_nameservers_search(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      nameservers:
        search:
          - mydomain.local
          - anotherdomain.local''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_nameservers_search', [])
        self.assertIn('mydomain.local', missing)
        self.assertIn('anotherdomain.local', missing)

    def test_diff_missing_netplan_nameservers_search(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: false
      dhcp6: false''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'dns_search': ['mydomain.local', 'anotherdomain.local'],
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_nameservers_search', [])
        self.assertIn('mydomain.local', missing)
        self.assertIn('anotherdomain.local', missing)

    def test_diff_nameservers_search_from_dhcp(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      dhcp4: true
      dhcp6: true''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'dns_search': ['mydomain.local', 'anotherdomain.local'],
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {
            'search': {
                'mydomain.local': 'DHCPv4',
                'anotherdomain.local': 'DHCPv6',
            }
        }
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_nameservers_search', [])
        self.assertEqual([], missing)

    def test_diff_missing_system_nameservers_search_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      nameservers:
        search:
          - mydomain.local
          - anotherdomain.local''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_nameservers_search', [])
        self.assertIn('mydomain.local', missing)
        self.assertIn('anotherdomain.local', missing)

    def test_diff_macaddress(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      macaddress: aa:bb:cc:dd:ee:ff''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'macaddress': '11:22:33:44:55:66'
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_macaddress')
        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_macaddress')
        self.assertEqual(missing_system, 'aa:bb:cc:dd:ee:ff')
        self.assertEqual(missing_netplan, '11:22:33:44:55:66')

    def test_diff_macaddress_option(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      macaddress: random''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'macaddress': '11:22:33:44:55:66'
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        # if the macaddress value in Netplan is one of the supported options, such as 'random', we don't calculated a diff
        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_macaddress')
        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_macaddress')
        self.assertEqual(missing_system, None)
        self.assertEqual(missing_netplan, None)

    def test_diff_macaddress_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      macaddress: aa:bb:cc:dd:ee:ff''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
                'macaddress': '11:22:33:44:55:66'
            }
        }

        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'mynic'
        interface1.data_sources = {}
        system_state.interface_list = [interface1]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_macaddress')
        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_macaddress')
        self.assertEqual(missing_system, 'aa:bb:cc:dd:ee:ff')
        self.assertEqual(missing_netplan, '11:22:33:44:55:66')

    def test__filter_system_routes_empty_inputs(self):
        filtered = self.diff_state._filter_system_routes(set(), [], {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_link_scope_routes(self):
        route = NetplanRoute(to='1.2.3.0/24', scope='link')
        filtered = self.diff_state._filter_system_routes({route}, [], {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_dhcp_ra_routes(self):
        route1 = NetplanRoute(protocol='dhcp')
        route2 = NetplanRoute(protocol='ra')
        filtered = self.diff_state._filter_system_routes({route1, route2}, [], {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_link_local_routes(self):
        route1 = NetplanRoute(scope='host', type='local', to='1.2.3.4', from_addr='1.2.3.4')
        # local 127.0.0.0/8 dev lo table local proto kernel scope host src 127.0.0.1
        route2 = NetplanRoute(scope='host', type='local', to='127.0.0.0/8', from_addr='127.0.0.1')
        system_addresses = ['1.2.3.4/24', '127.0.0.1/8']
        filtered = self.diff_state._filter_system_routes({route1, route2}, system_addresses, {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_keep_link_local_routes(self):
        route1 = NetplanRoute(scope='link', type='unicast', to='169.254.0.0/16', family=2, from_addr='169.254.65.85')
        route2 = NetplanRoute(scope='global', type='unicast', to='fe80::/64', family=10)
        system_addresses = []
        config = {'netplan_state': {'link_local': []}}
        filtered = self.diff_state._filter_system_routes({route1, route2}, system_addresses, config)
        # link local routes are present but they are disabled in the netdef
        self.assertSetEqual(filtered, {route1, route2})

    def test__filter_system_routes_remove_link_local_routes(self):
        route1 = NetplanRoute(scope='link', type='unicast', to='169.254.0.0/16', family=2, from_addr='169.254.65.85')
        route2 = NetplanRoute(scope='global', type='unicast', to='fe80::/64', family=10)
        system_addresses = []
        config = {'netplan_state': {'link_local': ['ipv4', 'ipv6']}}
        filtered = self.diff_state._filter_system_routes({route1, route2}, system_addresses, config)
        # link local routes are present and link local is enabled for both ipv4 and ipv6 in the netdef
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_link_local_routes_with_multiple_ips_same_subnet(self):
        # When an interface has multiple IPs in the same subnet the routing table will
        # have routes using one of the IPs as source. Example:
        # local 192.168.0.123 dev eth0 table local proto kernel scope host src 192.168.0.123
        # local 192.168.0.124 dev eth0 table local proto kernel scope host src 192.168.0.123
        route1 = NetplanRoute(scope='host', type='local', to='1.2.3.4', from_addr='1.2.3.4')
        route2 = NetplanRoute(scope='host', type='local', to='1.2.3.5', from_addr='1.2.3.4')
        system_addresses = ['1.2.3.4/24', '1.2.3.5/24']
        filtered = self.diff_state._filter_system_routes({route1, route2}, system_addresses, {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_ipv6_multicast_routes(self):
        route = NetplanRoute(type='multicast', to='ff00::/8', family=10)
        filtered = self.diff_state._filter_system_routes({route}, [], {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_ipv6_host_local_routes(self):
        route1 = NetplanRoute(family=10, to='fd42:bc43:e20e:8cf7:216:3eff:feaf:4121')
        route2 = NetplanRoute(family=10, to='fd42:bc43:e20e:8cf7::/64')
        addresses = ['fd42:bc43:e20e:8cf7:216:3eff:feaf:4121/64']
        filtered = self.diff_state._filter_system_routes({route1, route2}, addresses, {})
        self.assertSetEqual(filtered, set())

    def test__filter_system_routes_should_not_be_filtered(self):
        route1 = NetplanRoute(to='default', via='1.2.3.4')
        route2 = NetplanRoute(to='1.2.3.0/24', via='4.3.2.1')
        route3 = NetplanRoute(to='1:2:3::/64', via='1:2:3::1234')
        filtered = self.diff_state._filter_system_routes({route1, route2, route3}, [], {})
        self.assertSetEqual(filtered, {route1, route2, route3})

    def test_diff_missing_netplan_routes(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'addresses': [{'fd42:bc43:e20e:8cf7:216:3eff:feaf:4121': {'prefix': 64}}],
                'routes': [
                    {
                        'to': 'default',
                        'via': '192.168.5.1',
                        'from': '192.168.5.122',
                        'metric': 100,
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    },
                    {
                        'to': '192.168.5.0',
                        'via': '192.168.5.1',
                        'from': '192.168.5.122',
                        'type': 'unicast',
                        'scope': 'link',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    },
                    {
                        'to': '1.2.3.0/24',
                        'via': '192.168.5.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'dhcp',
                        'family': 2,
                        'table': 'main'
                    },
                    {
                        'to': 'abcd::/64',
                        'via': 'abcd::1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'ra',
                        'family': 10,
                        'table': 'main'
                    },
                    {
                        'to': 'fe80::/64',
                        'protocol': 'kernel',
                        'family': 10,
                        'table': 'main'
                    },
                    {
                        'type': 'multicast',
                        'to': 'ff00::/8',
                        'table': 'local',
                        'protocol': 'kernel',
                        'family': 10
                    },
                    {
                        'type': 'local',
                        'to': '10.86.126.148',
                        'table': 'local',
                        'protocol': 'kernel',
                        'scope': 'host',
                        'from': '10.86.126.148',
                        'family': 2
                    },
                    {
                        'type': 'local',
                        'to': 'fd42:bc43:e20e:8cf7:216:3eff:feaf:4121',
                        'table': 'local',
                        'protocol': 'kernel',
                        'scope': 'global',
                        'family': 10
                    }

                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        expected = {}
        expected['to'] = 'default'
        expected['via'] = '192.168.5.1'
        expected['from_addr'] = '192.168.5.122'
        expected['metric'] = 100
        expected['protocol'] = 'kernel'
        expected['family'] = 2
        expected['table'] = 254
        expected_route = NetplanRoute(**expected)

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_routes', [])
        self.assertIn(expected_route, missing)

    def test_diff_missing_system_routes(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      routes:
        - to: 1.2.3.0/24
          via: 192.168.0.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'routes': [
                    {
                        'to': 'default',
                        'via': '192.168.5.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        expected = {}
        expected['to'] = '1.2.3.0/24'
        expected['via'] = '192.168.0.1'
        expected['family'] = 2
        expected['table'] = 254
        expected_route = NetplanRoute(**expected)

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        self.assertEqual(expected_route, missing[0])

    def test_diff_missing_system_routes_with_match(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      routes:
        - to: 1.2.3.0/24
          via: 192.168.0.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
                'routes': [
                    {
                        'to': 'default',
                        'via': '192.168.5.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        expected = {}
        expected['to'] = '1.2.3.0/24'
        expected['via'] = '192.168.0.1'
        expected['family'] = 2
        expected['table'] = 254
        expected_route = NetplanRoute(**expected)

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        self.assertEqual(expected_route, missing[0])

    def test_diff_missing_system_routes_with_gateway46(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      gateway4: 192.168.0.1
      gateway6: abcd::1234
      routes:
        - to: 1.2.3.0/24
          via: 192.168.0.123''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
                'routes': [
                    {
                        'to': '1.2.3.0/24',
                        'via': '192.168.0.123',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        expected4 = {}
        expected4['to'] = 'default'
        expected4['via'] = '192.168.0.1'
        expected4['family'] = 2
        expected4['protocol'] = 'static'
        expected_route4 = NetplanRoute(**expected4)

        expected6 = {}
        expected6['to'] = 'default'
        expected6['via'] = 'abcd::1234'
        expected6['family'] = 10
        expected6['protocol'] = 'static'
        expected_route6 = NetplanRoute(**expected6)

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        self.assertSetEqual({expected_route4, expected_route6}, set(missing))

    def test_diff_routes_with_gateway46_no_diff(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    mynic:
      match:
        name: eth0
      gateway4: 192.168.0.1
      gateway6: abcd::1234
      routes:
        - to: 1.2.3.0/24
          via: 192.168.0.123''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'mynic',
                'index': 2,
                'routes': [
                    {
                        'to': '1.2.3.0/24',
                        'via': '192.168.0.123',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    },
                    {
                        'to': 'default',
                        'via': '192.168.0.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'static',
                        'family': 2,
                        'table': 'main',
                        'metric': 300
                    },
                    {
                        'to': 'default',
                        'via': 'abcd::1234',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'static',
                        'family': 10,
                        'table': 'main',
                        'metric': 300
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        self.assertListEqual(missing, [])

    def test_diff_slash_32_and_128_routes(self):
        ''' /32 and /128 route entries from "ip route show" will not have the prefix
            1.2.3.4 via 10.3.0.1 dev mainif proto static
            1:2:3::4 via 10:3::1 dev mainif proto static metric 1024 pref medium
        '''
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      routes:
        - to: 1:2:3::4/128
          via: 1:2:3::1
        - to: 1.2.3.4/32
          via: 192.168.0.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'routes': [
                    {
                        'to': '1.2.3.4',
                        'via': '192.168.0.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    },
                    {
                        'to': '1:2:3::4',
                        'via': '1:2:3::1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 10,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_routes', [])
        self.assertListEqual([], missing_system)
        self.assertListEqual([], missing_netplan)

    def test_diff_compressed_ipv6_routes(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      routes:
        - to: 1:2:3:0:0:0::4/64
          from: 1:2:3:0:0:0::123
          via: 1:2:3:0:0::1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'routes': [
                    {
                        'to': '1:2:3::4/64',
                        'via': '1:2:3::1',
                        'from': '1:2:3::123',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 10,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing_system = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_routes', [])
        missing_netplan = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_routes', [])
        self.assertListEqual([], missing_system)
        self.assertListEqual([], missing_netplan)

    def test_diff_json_encoder(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0:
      routes:
        - to: 1.2.3.0/24
          via: 192.168.0.1''')

        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'routes': [
                    {
                        'to': 'default',
                        'via': '192.168.5.1',
                        'type': 'unicast',
                        'scope': 'global',
                        'protocol': 'kernel',
                        'family': 2,
                        'table': 'main'
                    }
                ]
            }
        }
        system_state.interface_list = []

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        diff_data_str = json.dumps(diff_data, cls=DiffJSONEncoder)
        diff_data_dict = json.loads(diff_data_str)
        self.assertTrue(len(diff_data_dict['interfaces']['eth0']['system_state']['missing_routes']) > 0)
        self.assertTrue(len(diff_data_dict['interfaces']['eth0']['netplan_state']['missing_routes']) > 0)

    def test_diff_present_system_bond_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bonds:
    bond0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bond': 'bond0'
            },
            'bond0': {
                'name': 'bond0',
                'id': 'bond0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'bond0'
        interface2.netdef_id = 'bond0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_bond_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('bond0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_system_bond_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bonds:
    bond0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            },
            'bond0': {
                'name': 'bond0',
                'id': 'bond0',
                'index': 3,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'bond0'
        interface2.netdef_id = 'bond0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_bond_link')
        self.assertEqual(missing, 'bond0')

        missing = diff_data.get('interfaces', {}).get('bond0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])

    def test_diff_present_system_bridge_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bridges:
    br0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bridge': 'br0'
            },
            'br0': {
                'name': 'br0',
                'id': 'br0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'br0'
        interface2.netdef_id = 'br0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_bridge_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('br0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_system_bridge_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bridges:
    br0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            },
            'br0': {
                'name': 'br0',
                'id': 'br0',
                'index': 3,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'br0'
        interface2.netdef_id = 'br0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_bridge_link')
        self.assertEqual(missing, 'br0')

        missing = diff_data.get('interfaces', {}).get('br0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])

    def test_diff_present_system_vrf_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  vrfs:
    vrf0:
      table: 1000
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'vrf': 'vrf0'
            },
            'vrf0': {
                'name': 'vrf0',
                'id': 'vrf0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'vrf0'
        interface2.netdef_id = 'vrf0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_vrf_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('vrf0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_system_vrf_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  vrfs:
    vrf0:
      table: 1000
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
            },
            'vrf0': {
                'name': 'vrf0',
                'id': 'vrf0',
                'index': 3,
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'vrf0'
        interface2.netdef_id = 'vrf0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('system_state', {}).get('missing_vrf_link')
        self.assertEqual(missing, 'vrf0')

        missing = diff_data.get('interfaces', {}).get('vrf0', {}).get('system_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])

    def test_diff_present_netplan_bond_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bonds:
    bond0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bond': 'bond0'
            },
            'bond0': {
                'name': 'bond0',
                'id': 'bond0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'bond0'
        interface2.netdef_id = 'bond0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_bond_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('bond0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_netplan_bond_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
    eth1: {}
  bonds:
    bond0:
      interfaces:
        - eth1''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bond': 'bond0',
            },
            'eth1': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 3,
                'bond': 'bond0',
            },
            'bond0': {
                'name': 'bond0',
                'id': 'bond0',
                'index': 4,
                'interfaces': ['eth0', 'eth1']
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'eth1'
        interface2.netdef_id = 'eth1'
        interface2.data_sources = {}
        interface3 = Mock(spec=Interface)
        interface3.name = 'bond0'
        interface3.netdef_id = 'bond0'
        interface3.data_sources = {}
        system_state.interface_list = [interface1, interface2, interface3]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_bond_link')
        self.assertEqual(missing, 'bond0')

        missing = diff_data.get('interfaces', {}).get('bond0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])

    def test_diff_present_netplan_bridge_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  bridges:
    br0:
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bridge': 'br0'
            },
            'br0': {
                'name': 'br0',
                'id': 'br0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'br0'
        interface2.netdef_id = 'br0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_bridge_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('br0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_netplan_bridge_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
    eth1: {}
  bridges:
    br0:
      interfaces:
        - eth1''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'bridge': 'br0',
            },
            'eth1': {
                'name': 'eth1',
                'id': 'eth1',
                'index': 3,
                'bridge': 'br0',
            },
            'br0': {
                'name': 'br0',
                'id': 'br0',
                'index': 4,
                'interfaces': ['eth0', 'eth1']
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'eth1'
        interface2.netdef_id = 'eth1'
        interface2.data_sources = {}
        interface3 = Mock(spec=Interface)
        interface3.name = 'br0'
        interface3.netdef_id = 'br0'
        interface3.data_sources = {}
        system_state.interface_list = [interface1, interface2, interface3]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_bridge_link')
        self.assertEqual(missing, 'br0')

        missing = diff_data.get('interfaces', {}).get('br0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])

    def test_diff_present_netplan_vrf_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
  vrfs:
    vrf0:
      table: 1000
      interfaces:
        - eth0''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'vrf': 'vrf0'
            },
            'vrf0': {
                'name': 'vrf0',
                'id': 'vrf0',
                'index': 3,
                'interfaces': ['eth0'],
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'vrf0'
        interface2.netdef_id = 'vrf0'
        interface2.data_sources = {}
        system_state.interface_list = [interface1, interface2]

        diff = NetplanDiffState(system_state, netplan_state)
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_vrf_link')
        self.assertIsNone(missing)

        missing = diff_data.get('interfaces', {}).get('vrf0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, [])

    def test_diff_missing_netplan_vrf_link(self):
        with open(self.path, "w") as f:
            f.write('''network:
  ethernets:
    eth0: {}
    eth1: {}
  vrfs:
    vrf0:
      table: 1000
      interfaces:
        - eth1''')
        netplan_state = NetplanConfigState(rootdir=self.workdir.name)
        system_state = Mock(spec=SystemConfigState)

        system_state.get_data.return_value = {
            'netplan-global-state': {},
            'eth0': {
                'name': 'eth0',
                'id': 'eth0',
                'index': 2,
                'vrf': 'vrf0',
            },
            'eth1': {
                'name': 'eth1',
                'id': 'eth1',
                'index': 3,
                'vrf': 'vrf0',
            },
            'vrf0': {
                'name': 'vrf0',
                'id': 'vrf0',
                'index': 4,
                'interfaces': ['eth0', 'eth1']
            }
        }
        interface1 = Mock(spec=Interface)
        interface1.name = 'eth0'
        interface1.netdef_id = 'eth0'
        interface1.data_sources = {}
        interface2 = Mock(spec=Interface)
        interface2.name = 'eth1'
        interface2.netdef_id = 'eth1'
        interface2.data_sources = {}
        interface3 = Mock(spec=Interface)
        interface3.name = 'vrf0'
        interface3.netdef_id = 'vrf0'
        interface3.data_sources = {}
        system_state.interface_list = [interface1, interface2, interface3]

        diff = NetplanDiffState(system_state, netplan_state)
        diff.route_lookup_table_names = {
            0: 'unspec', 253: 'default', 254: 'main', 255: 'local',
            'unspec': 0, 'default': 253, 'main': 254, 'local': 255}
        diff_data = diff.get_diff()

        missing = diff_data.get('interfaces', {}).get('eth0', {}).get('netplan_state', {}).get('missing_vrf_link')
        self.assertEqual(missing, 'vrf0')

        missing = diff_data.get('interfaces', {}).get('vrf0', {}).get('netplan_state', {}).get('missing_interfaces', [])
        self.assertListEqual(missing, ['eth0'])



#!/usr/bin/python3
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan ip command line'''

import logging
import os
import sys
import subprocess
from subprocess import CalledProcessError

from .. import utils

lease_path = {
    'networkd': {
        'pattern': 'run/systemd/netif/leases/{lease_id}',
        'method': 'ifindex',
    },
    'NetworkManager': {
        'pattern': 'var/lib/NetworkManager/internal-{lease_id}-{interface}.lease',
        'method': 'nm_connection',
    },
}


class NetplanIp(utils.NetplanCommand):

    def __init__(self):
        super().__init__(command_id='ip',
                         description='Retrieve IP information from the system',
                         leaf=False)

    def run(self):
        self.command_leases = NetplanIpLeases()

        # subcommand: leases
        p_ip_leases = self.subparsers.add_parser('leases',
                                                 help='Display IP leases',
                                                 add_help=False)
        p_ip_leases.set_defaults(func=self.command_leases.run, commandclass=self.command_leases)

        self.parse_args()
        self.run_command()


class NetplanIpLeases(utils.NetplanCommand):

    def __init__(self):
        super().__init__(command_id='ip leases',
                         description='Display IP leases',
                         leaf=True)

    def run(self):
        self.parser.add_argument('interface',
                                 help='Interface for which to display IP lease settings.')
        self.parser.add_argument('--root-dir',
                                 help='Search for configuration files in this root directory instead of /')

        self.func = self.command_ip_leases

        self.parse_args()
        self.run_command()

    def command_ip_leases(self):

        if self.interface == 'help':  # pragma: nocover (covered in autopkgtest)
            self.print_usage()

        def find_lease_file(mapping):
            def lease_method_ifindex():
                ifindex_f = os.path.join('/sys/class/net', self.interface, 'ifindex')
                try:
                    with open(ifindex_f) as f:
                        return f.readlines()[0].strip()
                except Exception as e:
                    logging.debug('Cannot read file %s: %s', ifindex_f, str(e))
                    raise

            def lease_method_nm_connection():
                # FIXME: handle older versions of NM where 'nmcli dev show' doesn't exist
                try:
                    nmcli_dev_out = utils.nmcli_out(['dev', 'show', self.interface])
                    for line in nmcli_dev_out.splitlines():
                        if 'GENERAL.CONNECTION' in line:
                            conn_id = line.split(':')[1].rstrip().strip()
                            nmcli_con_out = utils.nmcli_out(['con', 'show', 'id', conn_id])
                            for line in nmcli_con_out.splitlines():
                                if 'connection.uuid' in line:
                                    return line.split(':')[1].rstrip().strip()
                except Exception as e:
                    raise Exception('Could not find a NetworkManager connection for the interface: %s' % str(e))
                raise Exception('Could not find a NetworkManager connection for the interface')

            lease_pattern = lease_path[mapping['backend']]['pattern']
            lease_method = lease_path[mapping['backend']]['method']

            try:
                lease_id = eval("lease_method_" + lease_method)()

                # We found something to build the path to the lease file with,
                # at this point we may have something to look at; but if not,
                # we'll rely on open() throwing an error.
                # This might happen if networkd doesn't use DHCP for the interface,
                # for instance.
                path = os.path.join('/',
                                    os.path.abspath(self.root_dir) if self.root_dir else "",
                                    lease_pattern.format(interface=self.interface,
                                                         lease_id=lease_id))
                # Fallback to 'dhclient' if no lease of NetworkManager's
                # internal DHCP client is found
                if not os.path.isfile(path):
                    path = path.replace('NetworkManager/internal-', 'NetworkManager/dhclient-')
                with open(path) as f:
                    for line in f.readlines():
                        print(line.rstrip())
            except Exception as e:
                print("No lease found for interface '%s': %s" % (self.interface, str(e)),
                      file=sys.stderr)
                sys.exit(1)

        # XXX: This code path is only still supported for legacy reasons and
        # not supposed to be called in the scope of a systemd generator. The
        # 'netplan status' command should be used instead. Should it be moved to
        # the 'configure' binary to keep legacy functionality around?
        # See also the comment in main() in src/generate.c.
        argv = [utils.get_generator_path()]
        if self.root_dir:
            argv += ['--root-dir', self.root_dir]
        argv += ['--mapping', self.interface]

        # Extract out of the generator our mapping in a dict.
        logging.debug('command ip leases: running %s', argv)
        try:
            out = subprocess.check_output(argv, text=True)
        except CalledProcessError:  # pragma: nocover (better be covered in autopkgtest)
            print("No lease found for interface '%s' (not managed by Netplan)" % self.interface, file=sys.stderr)
            sys.exit(1)
        mapping = {}
        mapping_s = out.split(',')
        for keyvalue in mapping_s:
            key, value = keyvalue.strip().split('=')
            mapping[key] = value

        find_lease_file(mapping)



#!/usr/bin/python3
#
# Copyright (C) 2018 Canonical, Ltd.
# Author: Mathieu Trudel-Lapierre <mathieu.trudel-lapierre@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''netplan configuration manager'''

import logging
import netplan
import os
import shutil
import sys
import tempfile

from typing import Optional


class ConfigManager(object):
    def __init__(self, prefix="/", extra_files={}):
        self.prefix = prefix
        self.extra_files = extra_files
        self.new_interfaces = set()
        self.np_state: Optional[netplan.State] = None

    def __getattr__(self, attr):
        if attr == '_tempdir':
            return getattr(super(), attr)
        assert self.np_state is not None, "Must call parse() before accessing the config."
        return getattr(self.np_state, attr)

    @property
    def physical_interfaces(self):
        assert self.np_state is not None, "Must call parse() before accessing the config."
        interfaces = {}
        interfaces.update(self.np_state.ethernets)
        interfaces.update(self.np_state.modems)
        interfaces.update(self.np_state.wifis)
        return interfaces

    @property
    def virtual_interfaces(self):
        assert self.np_state is not None, "Must call parse() before accessing the config."
        interfaces = {}
        # what about ovs_ports?
        interfaces.update(self.np_state.bridges)
        interfaces.update(self.np_state.bonds)
        interfaces.update(self.np_state.dummy_devices)
        interfaces.update(self.np_state.tunnels)
        interfaces.update(self.np_state.virtual_ethernets)
        interfaces.update(self.np_state.vlans)
        interfaces.update(self.np_state.vrfs)
        return interfaces

    def parse(self, extra_config=None):
        """
        Parse all our config files to return an object that describes the system's
        entire configuration, so that it can later be interrogated.

        Returns a libnetplan State wrapper
        """

        # /run/netplan shadows /etc/netplan/, which shadows /lib/netplan
        parser = netplan.Parser()
        try:
            parser.load_yaml_hierarchy(rootdir=self.prefix)

            if extra_config:
                for f in extra_config:
                    parser.load_yaml(f)

            self.np_state = netplan.State()
            self.np_state.import_parser_results(parser)
        except netplan.NetplanException as e:
            raise ConfigurationError(str(e))

        # Convoluted way to dump the parsed config to the logs...
        with tempfile.TemporaryFile() as tmp:
            self.np_state._dump_yaml(output_file=tmp)
            tmp.seek(0)
            logging.debug("Merged config:\n{}".format(tmp.read()))

        return self.np_state

    def add(self, config_dict):
        for config_file in config_dict:
            self._copy_file(config_file, config_dict[config_file])
        self.extra_files.update(config_dict)

        # Invalidate the current parsed state
        self.np_state = None

    @property
    def tempdir(self):
        # Moved out of __init__ to avoid creating tempdir if not needed
        # This broke configmanager on early boot when /tmp was not available
        if not hasattr(self, '_tempdir'):
            self._tempdir = tempfile.mkdtemp(prefix='netplan_')
        return self._tempdir

    def backup(self, backup_config_dir=True):
        temp_etc = os.path.join(self.tempdir, "etc")
        temp_run = os.path.join(self.tempdir, "run")
        if backup_config_dir:
            self.copy_tree(os.path.join(self.prefix, "etc/netplan"),
                           os.path.join(temp_etc, "netplan"))
        self.copy_tree(os.path.join(self.prefix, "run/NetworkManager/system-connections"),
                       os.path.join(temp_run, "NetworkManager", "system-connections"),
                       missing_ok=True)
        self.copy_tree(os.path.join(self.prefix, "run/systemd/network"),
                       os.path.join(temp_run, "systemd", "network"),
                       missing_ok=True)

    def revert(self):
        try:
            for extra_file in dict(self.extra_files):
                os.unlink(self.extra_files[extra_file])
                del self.extra_files[extra_file]
            temp_run = os.path.join(self.tempdir, "run")
            temp_nm_path = "{}/NetworkManager/system-connections".format(temp_run)
            temp_networkd_path = "{}/systemd/network".format(temp_run)
            if os.path.exists(temp_nm_path):
                shutil.rmtree(os.path.join(self.prefix, "run/NetworkManager/system-connections"))
                self.copy_tree(temp_nm_path,
                               os.path.join(self.prefix, "run/NetworkManager/system-connections"))
            if os.path.exists(temp_networkd_path):
                shutil.rmtree(os.path.join(self.prefix, "run/systemd/network"))
                self.copy_tree(temp_networkd_path,
                               os.path.join(self.prefix, "run/systemd/network"))
        except Exception as e:  # pragma: nocover (only relevant to filesystem failures)
            # If we reach here, we're in big trouble. We may have wiped out
            # file NM or networkd are using, and we most likely removed the
            # "new" config -- or at least our copy of it.
            # Given that we're in some halfway done revert; warn the user
            # aggressively and drop everything; leaving any remaining backups
            # around for the user to handle themselves.
            logging.error("Something really bad happened while reverting config: {}".format(e))
            logging.error("You should verify the netplan YAML in /etc/netplan and probably run 'netplan apply' again.")
            sys.exit(-1)

    def cleanup(self):
        if not hasattr(self, '_tempdir'):
            return
        shutil.rmtree(self.tempdir)

    def __del__(self):
        try:
            self.cleanup()
        except FileNotFoundError:
            # If cleanup() was called before, there is nothing to delete
            pass

    def _copy_file(self, src, dst):
        shutil.copy(src, dst)

    def copy_tree(self, src, dst, missing_ok=False, **kwargs):
        try:
            shutil.copytree(src, dst, copy_function=copy_with_ownership, **kwargs)
        except FileNotFoundError:
            if missing_ok:
                pass
            else:
                raise


def copy_with_ownership(src, dst, follow_symlinks=True):
    shutil.copy2(src, dst, follow_symlinks=follow_symlinks)
    stat = os.stat(src, follow_symlinks=follow_symlinks)
    os.chown(dst, stat.st_uid, stat.st_gid, follow_symlinks=follow_symlinks)


class ConfigurationError(Exception):
    """
    Configuration could not be parsed or has otherwise failed to apply
    """
    pass



#!/usr/bin/python3
# Functional tests of netplan CLI. These are run during "make check" and don't
# touch the system configuration at all.
#
# Copyright (C) 2016 Canonical, Ltd.
# Author: Martin Pitt <martin.pitt@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import subprocess
import unittest
import tempfile
import shutil
import yaml

from tests.test_utils import MockCmd
from generator.base import exe_generate

rootdir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
exe_cli = [os.path.join(rootdir, 'src', 'netplan.script')]
if shutil.which('python3-coverage'):
    exe_cli = ['python3-coverage', 'run', '--append', '--'] + exe_cli

os.environ.update({'LD_LIBRARY_PATH': '.:{}'.format(os.environ.get('LD_LIBRARY_PATH'))})

# combined the output of 'nmcli dev' and 'nmcli con' into a single mock output,
# connected via FAKE_NM_ID. Both are parsed line by line, checking for
# 'GENERAL.CONNECTION' or 'connection.id/uuid' respectively
nmcli_mock_output = '''
# nmcli dev ... MOCK OUTPUT:
GENERAL.CONNECTION:                     FAKE_NM_ID

# nmcli con ... MOCK OUTPUT:
connection.id:                          FAKE_NM_ID
connection.uuid:                        00000000-0000-0000-0000-000000000000
'''


def _load_yaml(text):
    return yaml.load(text, Loader=yaml.SafeLoader)


class TestArgs(unittest.TestCase):
    '''Generic argument parsing tests'''

    def test_global_help(self):
        out = subprocess.check_output(exe_cli + ['--help'])
        self.assertIn(b'Available commands', out)
        self.assertIn(b'generate', out)
        self.assertIn(b'--debug', out)

    def test_command_help(self):
        out = subprocess.check_output(exe_cli + ['generate', '--help'])
        self.assertIn(b'--root-dir', out)

    def test_no_command(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        p = subprocess.Popen(exe_cli, stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertIn(b'need to specify a command', err)
        self.assertNotEqual(p.returncode, 0)


class TestGenerate(unittest.TestCase):
    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory()
        generator = os.path.join(self.workdir.name, 'usr', 'lib', 'systemd',
                                 'system-generators', 'netplan')
        os.makedirs(os.path.dirname(generator))
        os.symlink(exe_generate, generator)

    def test_no_config(self):
        p = subprocess.Popen(exe_cli + ['generate', '--root-dir', self.workdir.name], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertEqual(set(os.listdir(self.workdir.name)), {'usr', 'run'})
        self.assertEqual(
            os.listdir(os.path.join(self.workdir.name, 'usr', 'lib', 'systemd', 'system-generators')),
            ['netplan'])
        self.assertEqual(
            os.listdir(os.path.join(self.workdir.name, 'run', 'systemd')),
            ['generator.late'])

    def test_with_empty_config(self):
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        path_b = os.path.join(c, 'b.yaml')
        open(path_a, 'w').close()
        with open(path_b, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    enlol: {dhcp4: yes}''')
        os.chmod(path_a, mode=0o600)
        os.chmod(path_b, mode=0o600)
        subprocess.check_call(exe_cli + ['generate', '--root-dir', self.workdir.name])
        self.assertEqual(os.listdir(os.path.join(self.workdir.name, 'run', 'systemd', 'network')),
                         ['10-netplan-enlol.network'])

    def test_with_config(self):
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    enlol: {dhcp4: yes}''')
        os.chmod(path_a, mode=0o600)
        out = subprocess.check_output(exe_cli + ['generate', '--root-dir', self.workdir.name])
        self.assertEqual(out, b'')
        self.assertEqual(os.listdir(os.path.join(self.workdir.name, 'run', 'systemd', 'network')),
                         ['10-netplan-enlol.network'])

    def test_mapping_for_unknown_iface(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    enlol: {dhcp4: yes}''')
        os.chmod(path_a, mode=0o600)
        p = subprocess.Popen(exe_cli +
                             ['generate', '--root-dir', self.workdir.name, '--mapping', 'nonexistent'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertNotEqual(p.returncode, 0)
        self.assertNotIn(b'nonexistent', out)

    def test_mapping_for_interface(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    enlol: {dhcp4: yes}''')
        os.chmod(path_a, mode=0o600)
        out = subprocess.check_output(exe_cli +
                                      ['generate', '--root-dir', self.workdir.name, '--mapping', 'enlol'])
        self.assertNotEqual(b'', out)
        self.assertIn('enlol', out.decode('utf-8'))

    def test_mapping_for_renamed_iface(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    myif:
      match:
        name: enlol
      set-name: renamediface
      dhcp4: yes
''')
        os.chmod(path_a, mode=0o600)
        out = subprocess.check_output(exe_cli +
                                      ['generate', '--root-dir', self.workdir.name, '--mapping', 'renamediface'])
        self.assertNotEqual(b'', out)
        self.assertIn('renamediface', out.decode('utf-8'))

    def test_netplan_try_ready_stamp_skip(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        os.environ.setdefault('NETPLAN_CONFIGURE_PATH', os.path.join(rootdir, 'configure'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  ethernets:
    myif:
      dhcp4: yes
''')
        os.chmod(path_a, mode=0o600)

        stamp_file = os.path.join(self.workdir.name, 'run', 'netplan', 'netplan-try.ready')
        os.makedirs(self.workdir.name + '/run/netplan', mode=0o700, exist_ok=True)
        open(stamp_file, 'w').close()  # create stamp file
        res = subprocess.run(exe_cli + ['--debug', 'generate', '--root-dir', self.workdir.name],
                             text=True, stderr=subprocess.PIPE)
        self.assertEqual(res.returncode, 1)
        self.assertIn('DEBUG:Skipping daemon-reload... \'netplan try\' is restoring configuration', res.stderr)


class TestIfupdownMigrate(unittest.TestCase):

    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory()
        self.ifaces_path = os.path.join(self.workdir.name, 'etc/network/interfaces')
        self.converted_path = os.path.join(self.workdir.name, 'etc/netplan/10-ifupdown.yaml')

    def test_system(self):
        os.environ.update({"ENABLE_TEST_COMMANDS": "1"})
        rc = subprocess.call(exe_cli + ['migrate', '--dry-run'],
                             stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        # may succeed or fail, but should not crash
        self.assertIn(rc, [0, 2])

    def do_test(self, iface_file, expect_success=True, dry_run=True, dropins=None):
        os.environ.update({"ENABLE_TEST_COMMANDS": "1"})
        if iface_file is not None:
            os.makedirs(os.path.dirname(self.ifaces_path))
            with open(self.ifaces_path, 'w') as f:
                f.write(iface_file)
        if dropins:
            for fname, contents in dropins.items():
                path = os.path.join(os.path.dirname(self.ifaces_path), fname)
                os.makedirs(os.path.dirname(path), exist_ok=True)
                with open(path, 'w') as f:
                    f.write(contents)

        argv = exe_cli + ['--debug', 'migrate', '--root-dir', self.workdir.name]
        if dry_run:
            argv.append('--dry-run')
        p = subprocess.Popen(argv, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        if expect_success:
            self.assertEqual(p.returncode, 0, err.decode())
        else:
            self.assertIn(p.returncode, [2, 3], err.decode())
        return (out, err)

    #
    # configs which can be converted
    #

    def test_no_config(self):
        (out, err) = self.do_test(None)
        self.assertEqual(out, b'')
        self.assertEqual(os.listdir(self.workdir.name), [])

    def test_only_empty_include(self):
        out = self.do_test('''# default interfaces file
source-directory /etc/network/interfaces.d''')[0]
        self.assertFalse(os.path.exists(self.converted_path))
        self.assertEqual(out, b'')

    def test_loopback_only(self):
        (out, err) = self.do_test('auto lo\n#ignore me\niface lo inet loopback')
        self.assertEqual(out, b'')
        self.assertIn(b'nothing to migrate\n', err)

    def test_dhcp4(self):
        out = self.do_test('auto en1\niface en1 inet dhcp')[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_dhcp6(self):
        out = self.do_test('auto en1\niface en1 inet6 dhcp')[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp6': True}}}}, out.decode())

    def test_dhcp4_and_6(self):
        out = self.do_test('auto lo\niface lo inet loopback\n\n'
                           'auto en1\niface en1 inet dhcp\niface en1 inet6 dhcp')[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True, 'dhcp6': True}}}}, out.decode())

    def test_includedir_rel(self):
        out = self.do_test('iface lo inet loopback\nauto lo\nsource-directory interfaces.d',
                           dropins={'interfaces.d/std': 'auto en1\niface en1 inet dhcp',
                                    'interfaces.d/std.bak': 'some_bogus dontreadme'})[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_includedir_abs(self):
        out = self.do_test('iface lo inet loopback\nauto lo\nsource-directory /etc/network/defs/my',
                           dropins={'defs/my/std': 'auto en1\niface en1 inet dhcp',
                                    'defs/my/std.bak': 'some_bogus dontreadme'})[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_include_rel(self):
        out = self.do_test('iface lo inet loopback\nauto lo\nsource interfaces.d/*.cfg',
                           dropins={'interfaces.d/std.cfg': 'auto en1\niface en1 inet dhcp',
                                    'interfaces.d/std.cfgold': 'some_bogus dontreadme'})[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_include_abs(self):
        out = self.do_test('iface lo inet loopback\nauto lo\nsource /etc/network/*.cfg',
                           dropins={'std.cfg': 'auto en1\niface en1 inet dhcp',
                                    'std.cfgold': 'some_bogus dontreadme'})[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_allow(self):
        out = self.do_test('allow-hotplug en1\niface en1 inet dhcp\n'
                           'allow-auto en2\niface en2 inet dhcp')[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True},
                          'en2': {'dhcp4': True}}}}, out.decode())

    def test_no_scripts(self):
        out = self.do_test('auto en1\niface en1 inet dhcp\nno-scripts en1')[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}}, out.decode())

    def test_write_file_noconfig(self):
        (out, err) = self.do_test('auto lo\niface lo inet loopback', dry_run=False)
        self.assertFalse(os.path.exists(self.converted_path))
        # should disable original ifupdown config
        self.assertFalse(os.path.exists(self.ifaces_path))
        self.assertTrue(os.path.exists(self.ifaces_path + '.netplan-converted'))

    def test_write_file_haveconfig(self):
        (out, err) = self.do_test('auto en1\niface en1 inet dhcp', dry_run=False)
        with open(self.converted_path) as f:
            config = _load_yaml(f)
        self.assertEqual(config, {'network': {
            'version': 2,
            'ethernets': {'en1': {'dhcp4': True}}}})

        # should disable original ifupdown config
        self.assertFalse(os.path.exists(self.ifaces_path))
        self.assertTrue(os.path.exists(self.ifaces_path + '.netplan-converted'))

    def test_write_file_prev_run(self):
        os.makedirs(os.path.dirname(self.converted_path))
        with open(self.converted_path, 'w') as f:
            f.write('canary')
        (out, err) = self.do_test('auto en1\niface en1 inet dhcp', dry_run=False, expect_success=False)
        with open(self.converted_path) as f:
            self.assertEqual(f.read(), 'canary')

        # should not disable original ifupdown config
        self.assertTrue(os.path.exists(self.ifaces_path))

    #
    # static
    #

    def test_static_ipv4_prefix(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8"]}}}}, out.decode())

    def test_static_ipv4_netmask(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4\nnetmask 255.0.0.0', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8"]}}}}, out.decode())

    def test_static_ipv4_alias(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\n'
                           'auto en1:1\niface en1:1 inet static\naddress 1.2.3.5/8', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8", "1.2.3.5/8"]}}}}, out.decode())

    def test_static_ipv4_no_address(self):
        out, err = self.do_test('auto en1\niface en1 inet static\nnetmask 1.2.3.4', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'no address supplied', err)

    def test_static_ipv4_no_network(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'does not specify prefix length, and netmask not specified', err)

    def test_static_ipv4_invalid_addr(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.400/8', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "1.2.3.400" as an IPv4 address', err)

    def test_static_ipv4_invalid_netmask(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4\nnetmask 123.123.123.0', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "1.2.3.4/123.123.123.0" as an IPv4 network', err)

    def test_static_ipv4_invalid_prefixlen(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/42', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "1.2.3.4/42" as an IPv4 network', err)

    def test_static_ipv4_unsupported_option(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/24\nmetric 1280', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'unsupported inet option "metric"', err)

    def test_static_ipv4_unknown_option(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/24\nxyzzy 1280', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'unknown inet option "xyzzy"', err)

    def test_static_ipv6_prefix(self):
        out = self.do_test('auto en1\niface en1 inet6 static\naddress fc00:0123:4567:89ab:cdef::1234/64', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["fc00:123:4567:89ab:cdef::1234/64"]}}}}, out.decode())

    def test_static_ipv6_netmask(self):
        out = self.do_test('auto en1\niface en1 inet6 static\n'
                           'address fc00:0123:4567:89ab:cdef::1234\nnetmask 64', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["fc00:123:4567:89ab:cdef::1234/64"]}}}}, out.decode())

    def test_static_ipv6_no_address(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\nnetmask 64', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'no address supplied', err)

    def test_static_ipv6_no_network(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'does not specify prefix length, and netmask not specified', err)

    def test_static_ipv6_invalid_addr(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::12345/64', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "fc00:0123:4567:89ab:cdef::12345" as an IPv6 address', err)

    def test_static_ipv6_invalid_netmask(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234\nnetmask 129', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "fc00:0123:4567:89ab:cdef::1234/129" as an IPv6 network', err)

    def test_static_ipv6_invalid_prefixlen(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234/129', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'error parsing "fc00:0123:4567:89ab:cdef::1234/129" as an IPv6 network', err)

    def test_static_ipv6_unsupported_option(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234/64\nmetric 1280', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'unsupported inet6 option "metric"', err)

    def test_static_ipv6_unknown_option(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234/64\nxyzzy 1280', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'unknown inet6 option "xyzzy"', err)

    def test_static_ipv6_accept_ra_0(self):
        out = self.do_test('auto en1\niface en1 inet6 static\n'
                           'address fc00:0123:4567:89ab:cdef::1234/64\naccept_ra 0', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["fc00:123:4567:89ab:cdef::1234/64"],
                                  'accept_ra': False}}}}, out.decode())

    def test_static_ipv6_accept_ra_1(self):
        out = self.do_test('auto en1\niface en1 inet6 static\n'
                           'address fc00:0123:4567:89ab:cdef::1234/64\naccept_ra 1', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["fc00:123:4567:89ab:cdef::1234/64"],
                                  'accept_ra': True}}}}, out.decode())

    def test_static_ipv6_accept_ra_2(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234/64\naccept_ra 2', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'netplan does not support accept_ra=2', err)

    def test_static_ipv6_accept_ra_unexpected(self):
        out, err = self.do_test('auto en1\niface en1 inet6 static\n'
                                'address fc00:0123:4567:89ab:cdef::1234/64\naccept_ra fish', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'unexpected accept_ra value "fish"', err)

    def test_static_gateway(self):
        out = self.do_test("""auto en1
iface en1 inet static
  address 1.2.3.4
  netmask 255.0.0.0
  gateway 1.1.1.1
iface en1 inet6 static
  address fc00:0123:4567:89ab:cdef::1234/64
  gateway fc00:0123:4567:89ab::1""", dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1':
                          {'addresses': ["1.2.3.4/8", "fc00:123:4567:89ab:cdef::1234/64"],
                           'gateway4': "1.1.1.1",
                           'gateway6': "fc00:0123:4567:89ab::1"}}}}, out.decode())

    def test_static_dns(self):
        out = self.do_test("""auto en1
iface en1 inet static
  address 1.2.3.4
  netmask 255.0.0.0
  dns-nameservers 1.2.1.1  1.2.2.1
  dns-search weird.network
iface en1 inet6 static
  address fc00:0123:4567:89ab:cdef::1234/64
  dns-nameservers fc00:0123:4567:89ab:1::1  fc00:0123:4567:89ab:2::1""", dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1':
                          {'addresses': ["1.2.3.4/8", "fc00:123:4567:89ab:cdef::1234/64"],
                           'nameservers': {
                               'search': ['weird.network'],
                               'addresses': ['1.2.1.1', '1.2.2.1',
                                             'fc00:0123:4567:89ab:1::1', 'fc00:0123:4567:89ab:2::1']
                           }}}}}, out.decode())

    def test_static_dns2(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\ndns-search foo  foo.bar', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8"],
                                  'nameservers': {
                                      'search': ['foo', 'foo.bar']
                                  }}}}}, out.decode())

    def test_static_mtu(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\nmtu 1280', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8"],
                                  'mtu': 1280}}}}, out.decode())

    def test_static_invalid_mtu(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\nmtu fish', expect_success=False)
        self.assertEqual(b'', out)
        self.assertIn(b'cannot parse "fish" as an MTU', err)

    def test_static_two_different_mtus(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\nmtu 1280\n'
                                'iface en1 inet6 static\naddress 2001::1/64\nmtu 9000', expect_success=False)
        self.assertEqual(b'', out)
        self.assertIn(b'tried to set MTU=9000, but already have MTU=1280', err)

    def test_static_hwaddress(self):
        out = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\nhwaddress 52:54:00:6b:3c:59', dry_run=True)[0]
        self.assertEqual(_load_yaml(out), {'network': {
            'version': 2,
            'ethernets': {'en1': {'addresses': ["1.2.3.4/8"],
                                  'macaddress': '52:54:00:6b:3c:59'}}}}, out.decode())

    def test_static_two_different_macs(self):
        out, err = self.do_test('auto en1\niface en1 inet static\naddress 1.2.3.4/8\nhwaddress 52:54:00:6b:3c:59\n'
                                'iface en1 inet6 static\naddress 2001::1/64\nhwaddress 52:54:00:6b:3c:58', expect_success=False)
        self.assertEqual(b'', out)
        self.assertIn(b'tried to set MAC 52:54:00:6b:3c:58, but already have MAC 52:54:00:6b:3c:59', err)

    #
    # configs which are not supported
    #

    def test_noauto(self):
        (out, err) = self.do_test('iface en1 inet dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'non-automatic interfaces are not supported', err)

    def test_dhcp_options(self):
        (out, err) = self.do_test('auto en1\niface en1 inet dhcp\nup myhook', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'option(s) up are not supported for dhcp method', err)

    def test_mapping(self):
        (out, err) = self.do_test('mapping en*\n  script /some/path/mapscheme\nmap HOME en1-home\n\n'
                                  'auto map1\niface map1 inet dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'mapping stanza is not supported', err)

    def test_unknown_allow(self):
        (out, err) = self.do_test('allow-foo en1\niface en1 inet dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Unknown stanza type allow-foo', err)

    def test_unknown_stanza(self):
        (out, err) = self.do_test('foo en1\niface en1 inet dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Unknown stanza type foo', err)

    def test_unknown_family(self):
        (out, err) = self.do_test('auto en1\niface en1 inet7 dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Unknown address family inet7', err)

    def test_unknown_method(self):
        (out, err) = self.do_test('auto en1\niface en1 inet mangle', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Unsupported method mangle', err)

    def test_too_few_fields(self):
        (out, err) = self.do_test('auto en1\niface en1 inet', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Expected 3 fields for stanza type iface but got 2', err)

    def test_too_many_fields(self):
        (out, err) = self.do_test('auto en1\niface en1 inet dhcp foo', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'Expected 3 fields for stanza type iface but got 4', err)

    def test_write_file_unsupported(self):
        (out, err) = self.do_test('iface en1 inet dhcp', expect_success=False)
        self.assertEqual(out, b'')
        self.assertIn(b'non-automatic interfaces are not supported', err)
        # should keep original ifupdown config
        self.assertTrue(os.path.exists(self.ifaces_path))


class TestInfo(unittest.TestCase):
    '''Test netplan info'''

    def test_info_defaults(self):
        """
        Check that 'netplan info' outputs at all, should include website URL
        """
        out = subprocess.check_output(exe_cli + ['info'])
        self.assertIn(b'features:', out)

    def test_info_yaml(self):
        """
        Verify that 'netplan info --yaml' output looks a bit like YAML
        """
        out = subprocess.check_output(exe_cli + ['info', '--yaml'])
        self.assertIn(b'features:', out)

    def test_info_json(self):
        """
        Verify that 'netplan info --json' output looks a bit like JSON
        """
        out = subprocess.check_output(exe_cli + ['info', '--json'])
        self.assertIn(b'"features": [', out)


class TestIp(unittest.TestCase):
    def setUp(self):
        self.workdir = tempfile.TemporaryDirectory()

    def test_valid_subcommand(self):
        p = subprocess.Popen(exe_cli + ['ip'], stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertIn(b'Available command', err)
        self.assertNotEqual(p.returncode, 0)

    def test_ip_leases_networkd(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            # match against loopback so as to successfully get a predictable
            # ifindex
            f.write('''network:
  version: 2
  renderer: networkd
  ethernets:
    enlol:
      match:
        name: lo
      dhcp4: yes
''')
        os.chmod(path_a, mode=0o600)
        fake_netif_lease_dir = os.path.join(self.workdir.name,
                                            'run', 'systemd', 'netif', 'leases')
        os.makedirs(fake_netif_lease_dir)
        with open(os.path.join(fake_netif_lease_dir, '1'), 'w') as f:
            f.write('''THIS IS A FAKE NETIF LEASE FOR LO''')
        out = subprocess.check_output(exe_cli +
                                      ['ip', 'leases',
                                       '--root-dir', self.workdir.name, 'lo'])
        self.assertNotEqual(out, b'')
        self.assertIn('FAKE NETIF', out.decode('utf-8'))

    def test_ip_leases_nm(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH',
                              os.path.join(rootdir, 'generate'))
        mock = MockCmd('nmcli')
        mock.set_output(nmcli_mock_output)
        new_path = os.pathsep.join([os.path.dirname(mock.path),
                                    os.environ.get('PATH', os.defpath)])
        mock_env = dict(os.environ, PATH=new_path)

        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  renderer: NetworkManager
  ethernets: {lo: {dhcp4: yes}}
''')
        os.chmod(path_a, mode=0o600)
        fake_lease_dir = os.path.join(self.workdir.name,
                                      'var', 'lib', 'NetworkManager')
        os.makedirs(fake_lease_dir)
        # using fake '0000..' UUID from nmcli_mock_output
        lease_file = 'internal-00000000-0000-0000-0000-000000000000-lo.lease'
        with open(os.path.join(fake_lease_dir, lease_file), 'w') as f:
            f.write('''THIS IS A FAKE INTERNAL_NM LEASE FOR LO''')
        out = subprocess.check_output(exe_cli +
                                      ['ip', 'leases',
                                       '--root-dir', self.workdir.name, 'lo'],
                                      env=mock_env)
        self.assertIn('FAKE INTERNAL_NM LEASE', out.decode('utf-8'))

    def test_ip_leases_nm_dhclient_fallback(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH',
                              os.path.join(rootdir, 'generate'))
        mock = MockCmd('nmcli')
        mock.set_output(nmcli_mock_output)
        new_path = os.pathsep.join([os.path.dirname(mock.path),
                                    os.environ.get('PATH', os.defpath)])
        mock_env = dict(os.environ, PATH=new_path)

        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  renderer: NetworkManager
  ethernets: {lo: {dhcp4: yes}}
''')
        os.chmod(path_a, mode=0o600)
        fake_lease_dir = os.path.join(self.workdir.name,
                                      'var', 'lib', 'NetworkManager')
        os.makedirs(fake_lease_dir)
        # using fake '0000..' UUID from nmcli_mock_output
        lease_file = 'dhclient-00000000-0000-0000-0000-000000000000-lo.lease'
        with open(os.path.join(fake_lease_dir, lease_file), 'w') as f:
            f.write('''THIS IS A FAKE DHCLIENT_NM LEASE FOR LO''')
        out = subprocess.check_output(exe_cli +
                                      ['ip', 'leases',
                                       '--root-dir', self.workdir.name, 'lo'],
                                      env=mock_env)
        self.assertIn('FAKE DHCLIENT_NM LEASE', out.decode('utf-8'))

    def test_ip_leases_nm_fail(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH',
                              os.path.join(rootdir, 'generate'))
        mock = MockCmd('nmcli')
        mock.set_output(nmcli_mock_output)
        mock.set_returncode(10)
        new_path = os.pathsep.join([os.path.dirname(mock.path),
                                    os.environ.get('PATH', os.defpath)])
        mock_env = dict(os.environ, PATH=new_path)

        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  renderer: NetworkManager
  ethernets: {lo: {dhcp4: yes}}
''')
        os.chmod(path_a, mode=0o600)
        # the nmcli Mock's return value is 10, indicating an error
        with self.assertRaises(Exception):
            subprocess.check_output(exe_cli +
                                    ['ip', 'leases',
                                     '--root-dir', self.workdir.name, 'lo'],
                                    env=mock_env)

    def test_ip_leases_no_networkd_lease(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            # match against loopback so as to successfully get a predictable
            # ifindex
            f.write('''network:
  version: 2
  ethernets:
    enlol:
      match:
        name: lo
      dhcp4: yes
''')
        os.chmod(path_a, mode=0o600)
        p = subprocess.Popen(exe_cli +
                             ['ip', 'leases', '--root-dir', self.workdir.name, 'enlol'],
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertIn(b'No lease found', err)
        self.assertNotEqual(p.returncode, 0)

    def test_ip_leases_no_nm_lease(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))
        mock = MockCmd('nmcli')
        mock.set_output('\n')
        new_path = os.pathsep.join([os.path.dirname(mock.path),
                                    os.environ.get('PATH', os.defpath)])
        mock_env = dict(os.environ, PATH=new_path)

        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        path_a = os.path.join(c, 'a.yaml')
        with open(path_a, 'w') as f:
            f.write('''network:
  version: 2
  renderer: NetworkManager
  ethernets:
    enlol:
      match:
        name: lo
      dhcp4: yes
''')
        os.chmod(path_a, mode=0o600)
        # we didn't create a (mock) lease file, therefore expect stderr output
        p = subprocess.Popen(exe_cli +
                             ['ip', 'leases', '--root-dir', self.workdir.name, 'enlol'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             env=mock_env)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertIn(b'No lease found', err)
        self.assertNotEqual(p.returncode, 0)

    def test_ip_leases_no_netdefs(self):
        os.environ.setdefault('NETPLAN_GENERATE_PATH', os.path.join(rootdir, 'generate'))

        c = os.path.join(self.workdir.name, 'etc', 'netplan')
        os.makedirs(c)
        with open(os.path.join(c, 'a.yaml'), 'w') as f:
            f.write('''network:
  version: 2
  renderer: NetworkManager
''')
        # we didn't create a (mock) lease file, therefore expect stderr output
        p = subprocess.Popen(exe_cli +
                             ['ip', 'leases', '--root-dir', self.workdir.name, 'enlol'],
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        (out, err) = p.communicate()
        self.assertEqual(out, b'')
        self.assertIn(b"No lease found for interface 'enlol' (not managed by Netplan)", err)
        self.assertNotEqual(p.returncode, 0)


unittest.main(testRunner=unittest.TextTestRunner(stream=sys.stdout, verbosity=2))
